import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.HashMap;

import ast.Assign;
import ast.BinOp;
import ast.Block;
import ast.Call;
import ast.Declaration;
import ast.Expression;
import ast.ExpressionStatement;
import ast.If;
import ast.InvalidExpression;
import ast.InvalidStatement;
import ast.Number;
import ast.Parameter;
import ast.Procedure;
import ast.Program;
import ast.Reference;
import ast.Referenceable;
import ast.Return;
import ast.Statement;
import ast.Type;
import ast.UnOp;
import ast.Visitor;
import ast.While;


/**
 * Export an ast to a graphviz dot file
 */
public class ExportAst {
	protected PrintWriter writer;
	protected String parentName;
	private int nextId;

	protected String makeNodeName() {
		return "node" + nextId++;
	}

	protected class DotExportVisitor implements Visitor {
		private HashMap<Referenceable, String> referenceableNames = new HashMap<Referenceable, String>();
		private HashMap<Procedure, String> procedureNames = new HashMap<Procedure, String>();

		private String getRefName(Referenceable referenceable) {
			String name = referenceableNames.get(referenceable);
			if (name == null) {
				name = makeNodeName();
				referenceableNames.put(referenceable, name);
			}
			return name;
		}

		private String getProcedureName(Procedure procedure) {
			String name = procedureNames.get(procedure);
			if (name == null) {
				name = makeNodeName();
				procedureNames.put(procedure, name);
			}
			return name;
		}

		@Override
		public void visit(Assign assign) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"= \\\"%s\\\"\",shape=diamond];\n", name, assign.getIdentifier());
			writer.printf("\t%s -> %s;\n", parentName, name);

			Referenceable destination = assign.getDestination();
			if (destination != null) {
				String destName = getRefName(destination);
				writer.printf("\t%s -> %s [style=dotted,color=blue]\n", name, destName);
			}

			String oldParent = parentName;
			parentName = name;
			assign.getExpression().accept(this);
			parentName = oldParent;
		}

		@Override
		public void visit(BinOp binOp) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"%s\",shape=diamond];\n", name, binOp.getBinOpType());
			writer.printf("\t%s -> %s;\n", parentName, name);

			String oldParent = parentName;
			parentName = name;
			binOp.getLeft().accept(this);
			binOp.getRight().accept(this);
			parentName = oldParent;
		}

		@Override
		public void visit(Block block) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"Block\",shape=box];\n", name);
			writer.printf("\t%s -> %s;\n", parentName, name);

			String oldParent = parentName;
			parentName = name;
			for (Statement statement : block.getStatements())
				statement.accept(this);
			parentName = oldParent;
		}

		@Override
		public void visit(Call call) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"Call \\\"%s\\\"\",shape=diamond];\n", name, call.getCallee());
			writer.printf("\t%s -> %s;\n", parentName, name);

			Procedure calledProcedure = call.getCalledProcedure();
			if (calledProcedure != null) {
				String destName = getProcedureName(calledProcedure);
				writer.printf("\t%s -> %s [style=dotted,color=blue];\n", name, destName);
			}

			String oldParent = parentName;
			parentName = name;
			for (Expression argument : call.getArguments())
				argument.accept(this);
			parentName = oldParent;
		}

		@Override
		public void visit(Declaration declaration) {
			String name = getRefName(declaration);
			writer.printf("\t%s [label=\"Decl \\\"%s\\\"\",shape=box];\n", name, declaration.getName());
			writer.printf("\t%s -> %s;\n", parentName, name);

			String oldParent = parentName;
			parentName = name;
			declaration.getType().accept(this);
			parentName = oldParent;
		}

		@Override
		public void visit(ExpressionStatement expressionStatement) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"ExpressionStatement\",shape=box];\n", name);
			writer.printf("\t%s -> %s;\n", parentName, name);

			String oldParent = parentName;
			parentName = name;
			expressionStatement.getExpression().accept(this);
			parentName = oldParent;
		}

		@Override
		public void visit(If ifs) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"If\",shape=box];\n", name);
			writer.printf("\t%s -> %s;\n", parentName, name);

			String oldParent = parentName;
			parentName = name;
			ifs.getCondition().accept(this);
			ifs.getThenStatement().accept(this);
			if (ifs.getElseStatement() != null)
				ifs.getElseStatement().accept(this);
			parentName = oldParent;
		}

		@Override
		public void visit(InvalidExpression invalidExpression) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"InvalidExpression\",shape=diamond,style=filled,color=red];\n", name);
			writer.printf("\t%s -> %s;\n", parentName, name);
		}

		@Override
		public void visit(InvalidStatement invalidStatement) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"InvalidStatement\",shape=diamond,style=filled,color=red];\n", name);
			writer.printf("\t%s -> %s;\n", parentName, name);
		}

		@Override
		public void visit(Number number) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"%s\",shape=diamond];\n", name, number.getString());
			writer.printf("\t%s -> %s;\n", parentName, name);
		}

		@Override
		public void visit(Parameter parameter) {
			String name = getRefName(parameter);
			writer.printf("\t%s [label=\"Param \\\"%s\\\"\",shape=box];\n", name, parameter.getName());
			writer.printf("\t%s -> %s;\n", parentName, name);

			String oldParent = parentName;
			parentName = name;
			parameter.getType().accept(this);
			parentName = oldParent;
		}

		@Override
		public void visit(Procedure procedure) {
			String name = getProcedureName(procedure);
			writer.printf("\t%s [label=\"Proc \\\"%s\\\"\",shape=box];\n", name, procedure.getName());
			writer.printf("\t%s -> %s;\n", parentName, name);

			String oldParent = parentName;
			parentName = name;
			procedure.getReturnType().accept(this);
			for (Parameter parameter : procedure.getParameters())
				parameter.accept(this);
			procedure.getStatement().accept(this);
			parentName = oldParent;
		}

		@Override
		public void visit(Program program) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"Program\",shape=box];\n", name);

			assert parentName == null;
			parentName = name;
			for (Procedure procedure : program.getProcedures()) {
				/* hide our builtin procedures... */
				if (procedure.getName().toString().equals("println")
						|| procedure.getName().toString().equals("read"))
					continue;
				procedure.accept(this);
			}
		}

		@Override
		public void visit(Reference reference) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"Ref \\\"%s\\\"\",shape=diamond];\n", name, reference.getIdentifier());
			writer.printf("\t%s -> %s;\n", parentName, name);

			Referenceable source = reference.getSource();
			if (source != null) {
				String sourceName = getRefName(source);
				writer.printf("\t%s -> %s [style=dotted,color=blue];\n", name, sourceName);
			}
		}

		@Override
		public void visit(Return returns) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"Return\",shape=box];\n", name);
			writer.printf("\t%s -> %s;\n", parentName, name);

			String oldParent = parentName;
			parentName = name;
			Expression expression = returns.getExpression();
			if (expression != null)
				expression.accept(this);
			parentName = oldParent;
		}

		@Override
		public void visit(Type type) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"%s\",shape=ellipse];\n", name, type);
			writer.printf("\t%s -> %s;\n", parentName, name);
		}

		@Override
		public void visit(UnOp unOp) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"%s\",shape=diamond];\n", name, unOp.getUnOpType());
			writer.printf("\t%s -> %s;\n", parentName, name);

			String oldParent = parentName;
			parentName = name;
			unOp.getOp().accept(this);
			parentName = oldParent;
		}

		@Override
		public void visit(While whiles) {
			String name = makeNodeName();
			writer.printf("\t%s [label=\"While\",shape=box];\n", name);
			writer.printf("\t%s -> %s;\n", parentName, name);

			String oldParent = parentName;
			parentName = name;
			whiles.getCondition().accept(this);
			whiles.getBody().accept(this);
			parentName = oldParent;
		}
	}

	private ExportAst(PrintWriter writer) {
		this.writer = writer;
	}

	private void doExport(Program program) {
		DotExportVisitor dotExportVisitor = new DotExportVisitor();
		writer.println("digraph Program {");

		program.accept(dotExportVisitor);

		writer.println("}");
	}

	public static void exportDot(Program program, PrintWriter writer) {
		ExportAst instance = new ExportAst(writer);
		instance.doExport(program);
	}

	public static void exportDot(Program program, String fileName) throws FileNotFoundException {
		PrintWriter writer = new PrintWriter(fileName);
		exportDot(program, writer);
		writer.close();
	}
}
