import java.io.IOException;
import java.io.Reader;
import java.util.Hashtable;

import nameana.StringTable;
import nameana.Symbol;
import tokens.Identifier;
import tokens.Number;
import tokens.SimpleToken;
import tokens.Token;
import tokens.TokenType;


public class Lexer {
	private Reader input;
	private StringTable stringTable;
	private int c;
	private Exception exception;
	private Hashtable<String, TokenType> simpleTokens = new Hashtable<String, TokenType>();

	public Lexer(Reader input, StringTable stringTable) {
		this.input = input;
		this.stringTable = stringTable;
		nextChar();
		simpleTokens.put("int", TokenType.INT);
		simpleTokens.put("boolean", TokenType.BOOLEAN);
		simpleTokens.put("void", TokenType.VOID);
		simpleTokens.put("if", TokenType.IF);
		simpleTokens.put("else", TokenType.ELSE);
		simpleTokens.put("while", TokenType.WHILE);
		simpleTokens.put("return", TokenType.RETURN);
	}

	private void nextChar() {
		try {
			c = input.read();
		} catch (IOException e) {
			c = -1;
			exception = e;
		}
	}

	public Token parseIdentifier() {
		StringBuilder string = new StringBuilder();
		do {
			string.append((char)c);
			nextChar();
		} while (Character.isUnicodeIdentifierPart(c));

		String result = string.toString();
		/* avoid StringTable */
		TokenType simpleTokenType = simpleTokens.get(result);
		if (simpleTokenType != null) {
			return new SimpleToken(simpleTokenType);
		}

		Symbol identifier = stringTable.get(result);
		return new Identifier(identifier);
	}

	public Token parseNumber() {
		StringBuilder string = new StringBuilder();
		do {
			string.append((char)c);
			nextChar();
		} while (Character.isDigit(c));
		return new Number(string.toString());
	}

	public Token next() {
		switch (c) {
		case -1:
			if (exception != null)
				return new SimpleToken(TokenType.ERROR);
			return new SimpleToken(TokenType.EOF);
		case ' ':
		case '\n':
		case '\r':
		case '\t':
			nextChar();
			return next();
		case '+': nextChar(); return new SimpleToken(TokenType.PLUS);
		case '-': nextChar(); return new SimpleToken(TokenType.MINUS);
		case '*': nextChar(); return new SimpleToken(TokenType.STAR);
		case '/': nextChar(); return new SimpleToken(TokenType.SLASH);
		case '(': nextChar(); return new SimpleToken(TokenType.LBRACE);
		case ')': nextChar(); return new SimpleToken(TokenType.RBRACE);
		case '{': nextChar(); return new SimpleToken(TokenType.LCURLY);
		case '}': nextChar(); return new SimpleToken(TokenType.RCURLY);
		case ',': nextChar(); return new SimpleToken(TokenType.COMMA);
		case ';': nextChar(); return new SimpleToken(TokenType.SEMICOLON);
		case '&':
			nextChar();
			if (c != '&') {
				exception = new Exception(String.format("Unexpected character '%s' after '&' while parsing", (char) c));
				return new SimpleToken(TokenType.ERROR);
			}
			nextChar();
			return new SimpleToken(TokenType.ANDAND);
		case '|':
			nextChar();
			if (c != '|') {
				exception = new Exception(String.format("Unexpected character '%s' after '|' while parsing", (char) c));
				return new SimpleToken(TokenType.ERROR);
			}
			nextChar();
			return new SimpleToken(TokenType.PIPEPIPE);
		case '!':
			nextChar();
			if (c == '=') {
				nextChar();
				return new SimpleToken(TokenType.EXCLAMATIONMARKEQUAL);
			}
			return new SimpleToken(TokenType.EXCLAMATIONMARK);
		case '=':
			nextChar();
			if (c == '=') {
				nextChar();
				return new SimpleToken(TokenType.EQUALEQUAL);
			}
			return new SimpleToken(TokenType.EQUAL);
		case '<':
			nextChar();
			if (c == '=') {
				nextChar();
				return new SimpleToken(TokenType.LESSEQUAL);
			}
			return new SimpleToken(TokenType.LESS);
		case '>':
			nextChar();
			if (c == '=') {
				nextChar();
				return new SimpleToken(TokenType.GREATEREQUAL);
			}
			return new SimpleToken(TokenType.GREATER);
		default:
			if (Character.isDigit(c)) {
				return parseNumber();
			}
			if (Character.isUnicodeIdentifierStart(c)) {
				return parseIdentifier();
			}
			exception = new Exception(String.format("Unexpected character '%s' while parsing", (char) c));
			return new SimpleToken(TokenType.ERROR);
		}
	}

	public Exception getException() {
		return exception;
	}
}