import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Reader;

import nameana.StringTable;
import tokens.Token;
import tokens.TokenType;
import ast.Program;

public class Main {
	private static String getFilenameWithoutExtension(String name) {
		/* search for '.' */
		int i;
		for (i = name.length()-1; i >= 0; --i)
			if (name.charAt(i) == '.')
				break;

		if (i < 0)
			return name;
		return name.substring(0, i);
	}

	private static void exportDot(String inputName, Program program) {
		try {
			String outputFileName = getFilenameWithoutExtension(inputName) + ".dot";
			ExportAst.exportDot(program, outputFileName);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	private static void lexerTest(Lexer lexer) {
		Token token;
		do {
			token = lexer.next();
			System.out.printf("Token: %s\n", token);
			if (token.getType() == TokenType.ERROR) {
				System.err.printf("Parse Error: %s\n", lexer.getException());
				System.exit(1);
			}
		} while (token.getType() != TokenType.EOF && token.getType() != TokenType.ERROR);
	}

	private static enum CompilationMode {
		NORMAL,
		LEXERTEST,
		PARSERTEST,
		SEMANTICTEST,
	}

	private static void printHelp() {
		System.out.println("Invocation:");
		System.out.println("\tMiniComp inputfile");
		System.out.println("\tMiniComp --lextest inputfile");
		System.out.println("\tMiniComp --parsetest inputfile");
		System.out.println("\tMiniComp --semantictest inputfile");
		System.out.println("\tMiniComp --help");
	}

	public static void main(String[] args) {
		boolean argumentProblem = false;
		CompilationMode mode = CompilationMode.NORMAL;
		String inputName = null;
		for (String argument : args) {
			argument = argument.intern(); // so we can use == for comparison
			if (argument == "--lextest")
				mode = CompilationMode.LEXERTEST;
			else if (argument == "--parsetest")
				mode = CompilationMode.PARSERTEST;
			else if (argument == "--semantictest")
				mode = CompilationMode.SEMANTICTEST;
			else if (argument == "--help") {
				printHelp();
				System.exit(0);
			} else if (argument.length() > 1 && argument.startsWith("-")) {
				System.err.printf("Error: Unknown option '%s'\n", argument);
				argumentProblem = true;
			} else {
				if (inputName != null) {
					System.err.println("Error: Multiple input files specified");
					argumentProblem = true;
				}
				inputName = argument;
			}
		}
		if (inputName == null) {
			System.err.println("Error: No inputfile specified");
			argumentProblem = true;
			inputName = "";
		}
		if (argumentProblem) {
			printHelp();
			System.exit(1);
		}

		Reader reader;
		if (inputName.equals("-")) {
			// input "-" is a shortcut for read from stdin
			reader = new InputStreamReader(System.in);
			inputName = "out"; // we use inputName to determine name of output files later
		} else {
			try {
				reader = new FileReader(inputName);
			} catch (FileNotFoundException e) {
				System.err.printf("Error: Could not open input file '%s'\n", inputName);
				System.exit(1);
				return;
			}
		}
		reader = new BufferedReader(reader);

		StringTable stringTable = new StringTable();
		Lexer lexer = new Lexer(reader, stringTable);
		if (mode == CompilationMode.LEXERTEST) {
			lexerTest(lexer);
			return;
		}

		Parser parser = new Parser(lexer, stringTable);
		Program program = parser.parseProgram();
		if (program == null) {
			System.err.println("Error: Input has invalid syntax, aborting.");
			System.exit(1);
		}
		if (mode == CompilationMode.PARSERTEST) {
			exportDot(inputName, program);
			return;
		}

		boolean semanticOk = Semantic.analyseSemantic(program);
		if (!semanticOk) {
			System.err.println("Error: Input had semantic errors, aborting.");
			System.exit(1);
		}
		if (mode == CompilationMode.SEMANTICTEST) {
			exportDot(inputName, program);
			return;
		}

		try {
			String noExtension = getFilenameWithoutExtension(inputName);
			String baseName = (new File(noExtension)).getName();
			String outputName =  baseName + ".j";
			PrintWriter writer = new PrintWriter(outputName);
			CodeGen.generateCode(writer, program, baseName);
			writer.close();

			/* assemble with jasmin */
			jasmin.Main.main(new String[] {outputName});
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

}
