import java.util.ArrayList;

import nameana.StringTable;
import nameana.Symbol;
import tokens.Identifier;
import tokens.Number;
import tokens.Token;
import tokens.TokenType;
import ast.Assign;
import ast.BinOp;
import ast.Block;
import ast.Call;
import ast.Declaration;
import ast.Expression;
import ast.ExpressionStatement;
import ast.If;
import ast.InvalidExpression;
import ast.InvalidStatement;
import ast.Parameter;
import ast.Procedure;
import ast.Program;
import ast.Return;
import ast.Statement;
import ast.Type;
import ast.UnOp;
import ast.While;

public class Parser {
	private Lexer lexer;
	private StringTable stringTable;
	private Token token;
	private boolean hadErrors;

	public Parser(Lexer lexer, StringTable stringTable) {
		this.lexer = lexer;
		this.stringTable = stringTable;
		nextToken();
	}
	private void nextToken() {
		token = lexer.next();
	}

	private void parseError(String message) {
		System.err.println(message);
		hadErrors = true;
	}

	private void skipTo(TokenType type) {
		while (token.getType() != type && token.getType() != TokenType.EOF)
			nextToken();
	}

	private void skip(TokenType type) {
		assert token.getType() == type;
		nextToken();
	}

	private Type parseType() {
		switch (token.getType()) {
		case INT:     nextToken(); return Type.INT;
		case BOOLEAN: nextToken(); return Type.BOOLEAN;
		case VOID:    nextToken(); return Type.VOID;
		default:
			parseError(String.format("Expected type, got token '%s'", token));
			return Type.INVALID;
		}
	}

	private Statement parseBlock() {
		Block result = new Block();
		skip(TokenType.LCURLY);
		while (token.getType() != TokenType.RCURLY) {
			if (token.getType() == TokenType.EOF) {
				parseError("Reached end-of-file, missing '}'");
				return result;
			}
			Statement statement = parseStatement();
			if (statement != null)
				result.addStatement(statement);
		}
		skip(TokenType.RCURLY);
		return result;
	}

	private Statement parseReturn() {
		skip(TokenType.RETURN);
		Expression result = null;
		if (token.getType() != TokenType.SEMICOLON) {
			result = parseExpression();
		}

		if (token.getType() != TokenType.SEMICOLON) {
			parseError(String.format("Expected semicolon after expression, got '%s'", token));
			skipTo(TokenType.SEMICOLON);
		}
		skip(TokenType.SEMICOLON);
		return new Return(result);
	}

	private Statement parseWhile() {
		skip(TokenType.WHILE);
		Expression condition = parseBraceExpression();
		Statement body = parseStatement();
		return new While(condition, body);
	}

	private Statement parseIf() {
		skip(TokenType.IF);
		Expression condition = parseBraceExpression();
		Statement thenStatement = parseStatement();
		Statement elseStatement = null;
		if (token.getType() == TokenType.ELSE) {
			skip(TokenType.ELSE);
			elseStatement = parseStatement();
		}
		return new If(condition, thenStatement, elseStatement);
	}

	private Statement parseExpressionStatement() {
		Expression expr = parseExpression();
		if (token.getType() != TokenType.SEMICOLON) {
			parseError(String.format("Expected semicolon after expression, got '%s'", token));
			skipTo(TokenType.SEMICOLON);
		}
		skip(TokenType.SEMICOLON);
		return new ExpressionStatement(expr);
	}

	private Statement parseDeclaration() {
		Type type = parseType();
		if (token.getType() != TokenType.IDENTIFIER) {
			parseError(String.format("Expected identifier, got '%s'", token));
			skipTo(TokenType.SEMICOLON);
			return new InvalidStatement();
		}
		Symbol name = ((Identifier) token).getSymbol();
		nextToken();
		if (token.getType() != TokenType.SEMICOLON) {
			parseError(String.format("Expected semicolon, got '%s'", token));
			skipTo(TokenType.SEMICOLON);
		} else {
			skip(TokenType.SEMICOLON);
		}
		return new Declaration(type, name);
	}

	private Statement parseStatement() {
		switch (token.getType()) {
		case LCURLY:
			return parseBlock();
		case WHILE:
			return parseWhile();
		case IF:
			return parseIf();
		case RETURN:
			return parseReturn();
		case INT:
		case BOOLEAN:
			return parseDeclaration();
		case NUMBER:
		case IDENTIFIER:
		case PLUS:
		case MINUS:
		case LBRACE:
			return parseExpressionStatement();
		case SEMICOLON:
			nextToken();
			return null;
		default:
			parseError(String.format("Expected statement, got token '%s'", token));
			return new InvalidStatement();
		}
	}

	private Expression parseExpression() {
		Expression expr = parseOrOp();
		while (true) {
			switch (token.getType()) {
			case PIPEPIPE: {
				skip(TokenType.PIPEPIPE);
				Expression right = parseOrOp();
				expr = new BinOp(BinOp.Type.LOGICALOR, expr, right);
				continue;
			}
			default:
				break;
			}
			break;
		}
		return expr;
	}

	private Expression parseOrOp() {
		Expression expr = parseAndOp();
		while (true) {
			switch (token.getType()) {
			case ANDAND: {
				skip(TokenType.ANDAND);
				Expression right = parseAndOp();
				expr = new BinOp(BinOp.Type.LOGICALAND, expr, right);
				continue;
			}
			default:
				break;
			}
			break;
		}
		return expr;
	}

	private Expression parseAndOp() {
		Expression expr = parseEqualityOp();
		while (true) {
			switch (token.getType()) {
			case EQUALEQUAL: {
				skip(TokenType.EQUALEQUAL);
				Expression right = parseEqualityOp();
				expr = new BinOp(BinOp.Type.EQUAL, expr, right);
				continue;
			}
			case EXCLAMATIONMARKEQUAL: {
				skip(TokenType.EXCLAMATIONMARKEQUAL);
				Expression right = parseEqualityOp();
				expr = new BinOp(BinOp.Type.NOTEQUAL, expr, right);
				continue;
			}
			default:
				break;
			}
			break;
		}
		return expr;
	}

	private Expression parseEqualityOp() {
		Expression expr = parseRelationalOp();
		while (true) {
			switch (token.getType()) {
			case LESS: {
				skip(TokenType.LESS);
				Expression right = parseRelationalOp();
				expr = new BinOp(BinOp.Type.LESS, expr, right);
				continue;
			}
			case LESSEQUAL: {
				skip(TokenType.LESSEQUAL);
				Expression right = parseRelationalOp();
				expr = new BinOp(BinOp.Type.LESSEQUAL, expr, right);
				continue;
			}
			case GREATER: {
				skip(TokenType.GREATER);
				Expression right = parseRelationalOp();
				expr = new BinOp(BinOp.Type.GREATER, expr, right);
				continue;
			}
			case GREATEREQUAL: {
				skip(TokenType.GREATEREQUAL);
				Expression right = parseRelationalOp();
				expr = new BinOp(BinOp.Type.GREATEREQUAL, expr, right);
				continue;
			}
			default:
				break;
			}
			break;
		}
		return expr;
	}

	private Expression parseRelationalOp() {
		Expression expr = parseTerm();
		while (true) {
			switch (token.getType()) {
			case PLUS: {
				skip(TokenType.PLUS);
				Expression right = parseTerm();
				expr = new BinOp(BinOp.Type.ADD, expr, right);
				continue;
			}
			case MINUS: {
				skip(TokenType.MINUS);
				Expression right = parseTerm();
				expr = new BinOp(BinOp.Type.SUB, expr, right);
				continue;
			}
			default:
				break;
			}
			break;
		}
		return expr;
	}

	private Expression parseTerm() {
		Expression expr = parseFactor();
		while (true) {
			switch (token.getType()) {
			case STAR: {
				skip(TokenType.STAR);
				Expression right = parseFactor();
				expr = new BinOp(BinOp.Type.MUL, expr, right);
				continue;
			}
			case SLASH: {
				skip(TokenType.SLASH);
				Expression right = parseFactor();
				expr = new BinOp(BinOp.Type.DIV, expr, right);
				continue;
			}
			default:
				break;
			}
			break;
		}
		return expr;
	}

	private Expression parseBraceExpression() {
		if (token.getType() != TokenType.LBRACE) {
			parseError(String.format("Expected '(expression)', got '%s'", token));
			return new InvalidExpression();
		}
		skip(TokenType.LBRACE);
		Expression result = parseExpression();
		if (token.getType() != TokenType.RBRACE) {
			parseError(String.format("Expected ')' got '%s'", token));
			skipTo(TokenType.SEMICOLON);
			return result;
		}
		skip(TokenType.RBRACE);
		return result;
	}

	private Expression parseCall(Symbol callee) {
		ArrayList<Expression> arguments = new ArrayList<Expression>();
		skip(TokenType.LBRACE);
		if (token.getType() != TokenType.RBRACE) {
			while (true) {
				if (token.getType() == TokenType.EOF) {
					parseError("Unexpecte end of file while parsing call arguments");
					return new InvalidExpression();
				}
				Expression argument = parseExpression();
				arguments.add(argument);
				if (token.getType() == TokenType.RBRACE) {
					break;
				}
				if (token.getType() != TokenType.COMMA) {
					parseError(String.format("Expected ')' or ',' after argument, got '%s'", token));
					skipTo(TokenType.RBRACE);
					break;
				}
				skip(TokenType.COMMA);
			}
		}
		skip(TokenType.RBRACE);
		return new Call(callee, arguments);
	}

	private Expression parseAssignment(Symbol destination) {
		skip(TokenType.EQUAL);
		Expression expression = parseExpression();
		return new Assign(destination, expression);
	}

	private Expression parseFactor() {
		switch (token.getType()) {
		case PLUS: // unary plus
			skip(TokenType.PLUS);
			return parseFactor();
		case MINUS: // unary minus
			skip(TokenType.MINUS);
			Expression op = parseFactor();
			return new UnOp(UnOp.Type.MINUS, op);
		case NUMBER: {
			Expression result = new ast.Number(((Number) token).getString());
			skip(TokenType.NUMBER);
			return result;
		}
		case IDENTIFIER: {
			Symbol symbol = ((Identifier) token).getSymbol();
			skip(TokenType.IDENTIFIER);
			if (token.getType() == TokenType.LBRACE) {
				return parseCall(symbol);
			}
			if (token.getType() == TokenType.EQUAL) {
				return parseAssignment(symbol);
			}
			Expression result = new ast.Reference(symbol);
			return result;
		}
		case LBRACE:
			return parseBraceExpression();
		default:
			parseError(String.format("Parse Error: Expected expression, got token '%s'", token));
			skipTo(TokenType.SEMICOLON);
			return new InvalidExpression();
		}
	}

	private ArrayList<Parameter> parseParameters() {
		ArrayList<Parameter> result = new ArrayList<Parameter>();
		if (token.getType() != TokenType.LBRACE) {
			parseError(String.format("expected parameterlist, got '%s'", token));
			skipTo(TokenType.LCURLY);
			return result;
		}
		skip(TokenType.LBRACE);
		boolean first = true;
		while (token.getType() != TokenType.RBRACE) {
			if (!first) {
				if (token.getType() != TokenType.COMMA) {
					parseError(String.format("Expected comma, got '%s'", token));
					skipTo(TokenType.RBRACE);
					return result;
				}
				skip(TokenType.COMMA);
			} else {
				first = false;
			}
			if (token.getType() == TokenType.EOF) {
				parseError("reached enf of file while parsing parameterlist");
				return result;
			}
			Type type = parseType();
			if (token.getType() != TokenType.IDENTIFIER) {
				parseError(String.format("Expected identifier, got '%s'", token));
				skipTo(TokenType.RBRACE);
				return result;
			}
			Symbol name = ((Identifier) token).getSymbol();
			skip(TokenType.IDENTIFIER);
			result.add(new Parameter(type, name));
		}
		skip(TokenType.RBRACE);
		return result;
	}

	private Procedure parseProcedure() {
		Type returnType = parseType();
		if (token.getType() != TokenType.IDENTIFIER) {
			parseError(String.format("Expected identifier, got '%s'", token));
			skipTo(TokenType.SEMICOLON);
			return null;
		}
		Symbol name = ((Identifier) token).getSymbol();
		nextToken();

		ArrayList<Parameter> parameters = parseParameters();
		if (token.getType() != TokenType.LCURLY) {
			parseError(String.format("Expected '{', got '%s'", token));
			return null;
		}
		Statement body = parseBlock();
		return new Procedure(returnType, name, body, parameters);
	}

	public Program parseProgram() {
		ArrayList<Procedure> procedures = new ArrayList<Procedure>();
		/* populate the procedure list with out 2 builtin functions
		 * read() and println()
		 */
		Symbol readSymbol = stringTable.get("read");
		Procedure read = new Procedure(Type.INT, readSymbol, null, new ArrayList<Parameter>());
		procedures.add(read);

		Symbol printlnSymbol = stringTable.get("println");
		ArrayList<Parameter> parameters = new ArrayList<Parameter>();
		parameters.add(new Parameter(Type.INT, stringTable.get("value")));
		Procedure println = new Procedure(Type.VOID, printlnSymbol, null, parameters);
		procedures.add(println);

		while (token.getType() != TokenType.EOF) {
			Procedure procedure = parseProcedure();
			procedures.add(procedure);
		}

		if (hadErrors)
			return null;
		return new Program(procedures);
	}
}