import java.util.ArrayList;
import java.util.HashMap;

import nameana.Symbol;
import nameana.SymbolTable;
import ast.Assign;
import ast.BinOp;
import ast.Block;
import ast.Call;
import ast.Declaration;
import ast.Expression;
import ast.ExpressionStatement;
import ast.If;
import ast.InvalidExpression;
import ast.InvalidStatement;
import ast.Number;
import ast.Parameter;
import ast.Procedure;
import ast.Program;
import ast.Reference;
import ast.Referenceable;
import ast.Return;
import ast.Statement;
import ast.Type;
import ast.UnOp;
import ast.Visitor;
import ast.While;


public class Semantic {
	private boolean hadErrors;

	protected void semanticError(String message) {
		System.err.println("Semantic error: " + message);
		hadErrors = true;
	}

	protected class SemanticCheckVisitor implements Visitor {
		private Procedure currentProcedure;
		private HashMap<Symbol, Procedure> procedures = new HashMap<Symbol, Procedure>();
		private SymbolTable symbolTable = new SymbolTable();

		@Override
		public void visit(Assign assign) {
			Symbol symbol = assign.getIdentifier();
			Referenceable destination = symbolTable.getCurrentDefinition(symbol);
			if (destination == null) {
				semanticError(String.format("Assignment to unknown identifier '%s'", symbol));
			} else {
				assign.setDestination(destination);
			}

			Expression expression = assign.getExpression();
			expression.accept(this);
			Type type = expression.getType();
			if (destination != null && !type.equals(destination.getType())) {
				semanticError(String.format("Wrong type in assignment to '%s', got %s, expected %s",
						symbol, type, destination.getType()));
				type = destination.getType();
			}
			assign.setType(type);
		}

		@Override
		public void visit(BinOp binOp) {
			Expression left = binOp.getLeft();
			Expression right = binOp.getRight();
			left.accept(this);
			right.accept(this);

			Type type = left.getType();
			if (!type.equals(right.getType())) {
				semanticError(String.format("Arguments to Binop(%s) have different types %s and %s",
						binOp.getBinOpType(), type, right.getType()));
			}
			switch (binOp.getBinOpType()) {
			case ADD:
			case MUL:
			case SUB:
			case DIV:
				if (!type.equals(Type.INT)) {
					semanticError(String.format("operator %s expects int operands, got %s", binOp.getBinOpType(), type));
					type = Type.INT;
				}
				break;
			case EQUAL:
			case NOTEQUAL:
			case LESS:
			case LESSEQUAL:
			case GREATER:
			case GREATEREQUAL:
				if (!type.equals(Type.INT)) {
					semanticError(String.format("operator %s expects int operands, got %s", binOp.getBinOpType(), type));
				}
				type = Type.BOOLEAN;
				break;
			case LOGICALAND:
			case LOGICALOR:
				if (!type.equals(Type.BOOLEAN)) {
					semanticError(String.format("operator %s expects boolean operands, got %s", binOp.getBinOpType(), type));
				}
				type = Type.BOOLEAN;
				break;
			}
			binOp.setType(type);
		}

		@Override
		public void visit(Block block) {
			symbolTable.enterScope();

			for (Statement statement : block.getStatements())
				statement.accept(this);

			symbolTable.leaveScope();
		}

		@Override
		public void visit(Call call) {
			Symbol callee = call.getCallee();
			Procedure called = procedures.get(callee);
			if (called == null) {
				semanticError(String.format("Called function '%s' is unknown", callee));
				return;
			}

			call.setCalledProcedure(called);
			call.setType(called.getReturnType());

			ArrayList<Expression> arguments = call.getArguments();
			ArrayList<Parameter> parameters = called.getParameters();
			if (arguments.size() != parameters.size()) {
				semanticError(String.format("Wrong number of arguments %d when calling '%s', expected %d",
						arguments.size(), callee, parameters.size()));
				return;
			}

			for (int i = 0; i < arguments.size(); ++i) {
				Expression argument = arguments.get(i);
				Parameter parameter = parameters.get(i);

				argument.accept(this);
				if (!argument.getType().equals(argument.getType())) {
					semanticError(String.format("Type mismatch in parameter %d when calling '%s': %s, expected %s",
							i, callee, argument.getType(), parameter.getType()));
				}
			}
		}

		@Override
		public void visit(Declaration declaration) {
			Symbol symbol = declaration.getName();
			if (symbolTable.definitionInCurrentScope(symbol)) {
				semanticError(String.format("Symbol '%s' defined twice", symbol));
			}
			symbolTable.enterDefinition(symbol, declaration);

			/* codegenerator cannot handle this yet */
			if (declaration.getType().equals(Type.BOOLEAN)) {
				semanticError(String.format("Variable '%s' must not have type boolean", symbol));
			}
			if (declaration.getType().equals(Type.VOID)) {
				semanticError(String.format("Variable '%s' must not have type void", symbol));
			}
		}

		@Override
		public void visit(ExpressionStatement expressionStatement) {
			expressionStatement.getExpression().accept(this);
		}

		@Override
		public void visit(If ifs) {
			Expression condition = ifs.getCondition();
			condition.accept(this);
			if (!condition.getType().equals(Type.BOOLEAN)) {
				semanticError(String.format("Condition '%s' in if-statement has not type boolean but %s",
						condition, condition.getType()));
			}
			ifs.getThenStatement().accept(this);
			if (ifs.getElseStatement() != null)
				ifs.getElseStatement().accept(this);
		}

		@Override
		public void visit(InvalidExpression invalidExpression) {
			/* should not happen */
			throw new RuntimeException("Internal compiler error");
		}

		@Override
		public void visit(InvalidStatement invalidStatement) {
			/* should not happen */
			throw new RuntimeException("Internal compiler error");
		}

		@Override
		public void visit(Number number) {
			number.setType(Type.INT);
		}

		@Override
		public void visit(Parameter parameter) {
			Symbol symbol = parameter.getName();
			if (symbolTable.definitionInCurrentScope(symbol)) {
				semanticError(String.format("Multiple parameters with name '%s'", symbol));
			}
			symbolTable.enterDefinition(symbol, parameter);

			/* codegenerator cannot handle this yet */
			if (parameter.getType().equals(Type.BOOLEAN)) {
				semanticError(String.format("Variable '%s' must not have type boolean", symbol));
			}
			if (parameter.getType().equals(Type.VOID)) {
				semanticError(String.format("Parameter '%s' must not have type VOID", symbol));
			}
		}

		@Override
		public void visit(Procedure procedure) {
			/* codegenerator cannot handle this yet */
			if (procedure.getReturnType().equals(Type.BOOLEAN)) {
				semanticError(String.format("Procedure '%s' must not have return type boolean",
						procedure.getName()));
			}

			symbolTable.enterScope();
			for (Parameter parameter : procedure.getParameters())
				parameter.accept(this);
			Statement statement = procedure.getStatement();
			/* builtin procedure have no statement */
			if (statement != null)
				statement.accept(this);
			symbolTable.leaveScope();
		}

		@Override
		public void visit(Program program) {
			/* pass1: collect Procedures */
			for (Procedure procedure : program.getProcedures()) {
				currentProcedure = procedure;
				procedures.put(procedure.getName(), procedure);
			}

			/* pass2: check bodies */
			for (Procedure procedure : program.getProcedures()) {
				currentProcedure = procedure;
				procedure.accept(this);
			}
		}

		@Override
		public void visit(Reference reference) {
			Symbol symbol = reference.getIdentifier();
			Referenceable source = symbolTable.getCurrentDefinition(symbol);
			if (source == null) {
				semanticError(String.format("Reference to unknown identifier '%s'", symbol));
				reference.setType(Type.INVALID);
				return;
			}

			reference.setSource(source);
			reference.setType(source.getType());
		}

		@Override
		public void visit(Return returns) {
			Expression expression = returns.getExpression();
			if (expression != null) {
				expression.accept(this);
				if (!expression.getType().equals(currentProcedure.getReturnType())) {
					semanticError(String.format("Type of return-expression is %s, expected %s",
							expression.getType(), currentProcedure.getReturnType()));
				}
			} else {
				if (!currentProcedure.getReturnType().equals(Type.VOID)) {
					semanticError(String.format("return without expression in procedure returning %s",
							currentProcedure.getReturnType()));
				}
			}
		}

		@Override
		public void visit(Type type) {
			/* nothing to do */
		}

		@Override
		public void visit(UnOp unOp) {
			Expression op = unOp.getOp();
			op.accept(this);
			Type type = op.getType();
			switch (unOp.getUnOpType()) {
			case MINUS:
				if (!type.equals(Type.INT)) {
					semanticError(String.format("- operator expects type int, got %s", type));
					type = Type.INT;
				}
				break;
			case NOT:
				if (!type.equals(Type.BOOLEAN)) {
					semanticError(String.format("! operator expects type boolean, got %s", type));
					type = Type.BOOLEAN;
				}
				break;
			}
			unOp.setType(type);
		}

		@Override
		public void visit(While whiles) {
			Expression condition = whiles.getCondition();
			condition.accept(this);
			if (!condition.getType().equals(Type.BOOLEAN)) {
				semanticError(String.format("Condition '%s' in while-statement has not type boolean but %s",
						condition, condition.getType()));
			}
			whiles.getBody().accept(this);
		}
	}

	private Semantic() {
		/* empty */
	}

	private boolean doAnalysis(Program program) {
		SemanticCheckVisitor visitor = new SemanticCheckVisitor();
		program.accept(visitor);
		return !hadErrors;
	}

	public static boolean analyseSemantic(Program program) {
		Semantic semantic = new Semantic();
		return semantic.doAnalysis(program);
	}
}
